import json
from typing import Dict, Final

import requests
from requests.models import HTTPError, Response

from nordigen.api import (
    AccountApi,
    AgreementsApi,
    InstitutionsApi,
    RequisitionsApi,
)
from nordigen.types.http_enums import HTTPMethod
from nordigen.types.types import RequisitionDto, TokenType


class NordigenClient:
    """
    Class to initialize new Client.

    Attributes
    ---------
    secret_key (str): Generated secret_key
    secret_id (str): Generated secret_id
    """

    __ENDPOINT: Final = "token"

    def __init__(
        self,
        secret_key: str,
        secret_id: str,
    ) -> None:
        self.secret_key = secret_key
        self.secret_id = secret_id
        self.base_url = "https://ob.nordigen.com/api/v2"
        self._headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "Nordigen-Python-v2",
        }
        self._token = None
        self.institution = InstitutionsApi(client=self)
        self.requisition = RequisitionsApi(client=self)
        self.agreement = AgreementsApi(client=self)

    def account_api(self, id: str) -> AccountApi:
        """
        Create Account api instance.

        Args:
            id (str): account id

        Returns:
            AccountApi: Account instance
        """
        return AccountApi(client=self, id=id)

    @property
    def token(self):
        """
        Get token.

        Returns:
            str: return token
        """
        return self._token

    @token.setter
    def token(self, value: str):
        """
        Set token.

        Args:
            value (str): token
        """
        self._token = value
        self._headers["Authorization"] = f"Bearer {value}"

    def generate_token(self) -> TokenType:
        """
        Generate new access token.

        Returns:
            TokenType: Dict that contains access and refresh token
        """
        payload = {"secret_key": self.secret_key, "secret_id": self.secret_id}
        response = self.request(
            HTTPMethod.POST,
            f"{self.__ENDPOINT}/new/",
            payload,
            self._headers,
        )
        self._headers["Authorization"] = f"Bearer {response['access']}"
        return response

    def exchange_token(self, refresh_token: str) -> TokenType:
        """
        Exchange refresh token for access token.

        Args:
            refresh_token (str): refresh token

        Returns:
            TokenType: Dict that contains access and refresh token
        """
        payload = {"refresh": refresh_token}
        return self.request(
            HTTPMethod.POST,
            f"{self.__ENDPOINT}/refresh/",
            payload,
            self._headers,
        )

    def request(
        self,
        method: HTTPMethod,
        endpoint: str,
        data: Dict = None,
        headers: Dict = None,
    ) -> Response:
        """
        Request wrapper for Nordigen library.

        Args:
            method (HTTPMethod): Supports GET, POST, PUT, DELETE
            endpoint (str): [endpoint url
            data (Dict, optional): body or parameters that need to be sent alongside with the request.
                Defaults to {}.

        Raises:
            Exception: HTTP method is not supported
            HTTPError: HTTP error with status code

        Returns:
            Response: JSON Response object
        """
        request_meta = {
            "url": f"{self.base_url}/{endpoint}",
            "headers": headers if headers else self._headers,
        }
        if method == HTTPMethod.GET:
            response = requests.get(**request_meta, params=data)
        elif method == HTTPMethod.POST:
            response = requests.post(**request_meta, data=json.dumps(data))
        elif method == HTTPMethod.PUT:
            response = requests.put(**request_meta, data=json.dumps(data))
        elif method == HTTPMethod.DELETE:
            response = requests.delete(**request_meta, params=data)
        else:
            raise Exception(f'Method "{method}" is not supported')

        if response.ok:
            return response.json()

        raise HTTPError(
            {"response": response.json(), "status": response.status_code}
        )

    def initialize_session(
        self,
        redirect_uri: str,
        institution_id: str,
        reference_id: str,
        max_historical_days: int = 90,
        access_valid_for_days: int = 90,
    ) -> RequisitionDto:
        """
        Factory method that creates authorization in a specific institution
        and are responsible for the following steps:
            * Creates agreement
            * Creates requisition

        Returns:
            Dict[str]: link to initiate authorization with bank and requisition_id
        """
        # Create agreement
        agreement = self.agreement.create_agreement(
            max_historical_days=max_historical_days,
            access_valid_for_days=access_valid_for_days,
            institution_id=institution_id,
        )

        requisition_dict = {
            "redirect_uri": redirect_uri,
            "reference_id": reference_id,
            "institution_id": institution_id,
            "agreement": agreement["id"],
        }

        # Create requisition
        requisition = self.requisition.create_requisition(**requisition_dict)

        return RequisitionDto(
            link=requisition["link"], requisition_id=requisition["id"]
        )



from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional, TypedDict


class TokenType(TypedDict):
    access: str
    refresh: str
    access_expires: int
    refresh_expires: int


class Institutions(TypedDict):
    id: str
    name: str
    bic: str
    transaction_total_days: str
    countries: List[str]
    logo: str


class EnduserAgreement(TypedDict):
    id: str
    created: str
    enduser_id: str
    institution_id: str
    accepted: Optional[str]
    access_scope: List[str]
    max_historical_days: int
    access_valid_for_days: int


class AgreementsList(TypedDict):
    count: int
    next: Optional[str]
    previous: Optional[str]
    results: List[EnduserAgreement]


class Requisition(TypedDict):
    id: str
    created: str
    redirect: str
    status: str
    agreements: List[str]
    accounts: List[str]
    reference: str
    enduser_id: str
    user_language: Optional[str]
    results: List[Dict[str, str]]


class RequisitionList(AgreementsList):
    results: List[Requisition]


@dataclass
class RequisitionDto:
    link: str
    requisition_id: str


class Balances(TypedDict):
    amount: str
    currency: str


class AccountBalances(TypedDict):
    balances: Balances
    balanceType: str
    creditLimitIncluded: Optional[bool]
    lastChangeDateTime: Optional[datetime]
    referenceDate: Optional[datetime]
    lastCommittedTransaction: Optional[str]


class AccountData(TypedDict):
    id: str
    created: datetime
    lastAccessed: datetime
    iban: str
    aspspIdentifier: str
    status: str


class AccountInfo(TypedDict):
    resource_id: str
    iban: str
    currency: str
    ownerName: str


class AccountDetails(TypedDict):
    account: AccountInfo



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Final, List, Union

from nordigen.types.http_enums import HTTPMethod
from nordigen.types.types import AgreementsList, EnduserAgreement

if TYPE_CHECKING:
    from nordigen.client import NordigenClient


class AgreementsApi:
    """
    Agreements api class.

    Attributes
    -------
    client (NordigenClient): Injectable NordigenClient object to make an http requests
    institution_id (str): institution_id (bank id) from institutions response.
        e.g {"id": "MONZO_MONZGB2L"}

    Returns: None
    """

    __ENDPOINT: Final = "agreements/enduser"

    def __init__(self, client: NordigenClient) -> None:
        self.__client = client

    def create_agreement(
        self,
        institution_id: str,
        max_historical_days: int = 90,
        access_valid_for_days: int = 90,
        access_scope: Union[None, List[str]] = None,
    ) -> EnduserAgreement:
        """
        Create end user agreement.

        Args:
            institution (str): Institution id.
            max_historical_days (int, optional): Length of the transaction history. Defaults to 90.
            access_valid_for_days (int, optional): access valid for days.
            access_scope (List[str], optional):
                access scope for account, by default provides access to balances, details, transactions.
                Defaults to [ "balances", "details", "transactions" ].

        Returns:
            EnduserAgreement: Enduser agreement json object
        """
        access_scope = access_scope or ["balances", "details", "transactions"]
        payload = {
            "max_historical_days": max_historical_days,
            "access_valid_for_days": access_valid_for_days,
            "access_scope": access_scope,
            "institution_id": institution_id,
        }

        return self.__client.request(
            HTTPMethod.POST, f"{self.__ENDPOINT}/", payload
        )

    def get_agreements(
        self, limit: int = 100, offset: int = 0
    ) -> AgreementsList:
        """
        Get list of agreements.

        Args:
                 a unique end-user ID of someone who's using your services, it has to be unique within your solution.
                 Usually, it's UUID
            limit (int, optional): number of results to return per page. Defaults to 100.
            offset (int, optional): the initial index from which to return the results. Defaults to 0.

        Returns:
            AgreementsList: json object with enduser agreements
        """
        params = {"limit": limit, "offset": offset}
        return self.__client.request(
            HTTPMethod.GET, f"{self.__ENDPOINT}/", params
        )

    def get_agreement_by_id(self, agreement_id: str) -> EnduserAgreement:
        """
        Get agreement by agreement id.

        Args:
            agreement_id (str): id id from create_agreement response
        Returns:
            EnduserAgreement: JSON object with specific enduser agreements
        """
        return self.__client.request(
            HTTPMethod.GET, f"{self.__ENDPOINT}/{agreement_id}"
        )

    def delete_agreement(self, agreement_id: str) -> Dict:
        """
        Delete End User Agreement by id.

        Args:
            agreement_id (str): A UUID string identifying this end user agreement.

        Returns:
            Dict: Dictionary with deleted agreement
        """
        return self.__client.request(
            HTTPMethod.DELETE, f"{self.__ENDPOINT}/{agreement_id}"
        )

    def accept_agreement(
        self, agreement_id: str, ip: str, user_agent: str
    ) -> Dict:
        """
        Accept an end-user agreement via the API.

        Args:
            agreement_id (str): A UUID string identifying this end user agreement.
            ip (str): IP address of the client
            user_agent (str): [User Agent of the browser
        Returns:
            Dict: Dict with information on accepted agreement
        """
        payload = {"user_agent": user_agent, "ip_address": ip}
        return self.__client.request(
            HTTPMethod.PUT,
            f"{self.__ENDPOINT}/{agreement_id}/accept/",
            payload,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Final, List

from nordigen.types import Institutions
from nordigen.types.http_enums import HTTPMethod

if TYPE_CHECKING:
    from nordigen import NordigenClient


class InstitutionsApi:
    """
    Institution API class. Used to fetch information about bank.

    Attributes
    ---------
    client(NordigenClient): Injectable NordigenClient object to make an http requests

    Returns: None
    """

    ENDPOINT: Final = "aspsps"

    def __init__(self, client: NordigenClient) -> None:
        self.__request = client.request
        self.ENDPOINT = "institutions"

    def get_institutions(self, country: str) -> List[Institutions]:
        """
        Get all available Institutions (banks) in a given country.

        Args:
            country (str): Two-character country code

        Returns:
            List[Institutions]: List of institutions in a given country
        """
        return self.__request(
            HTTPMethod.GET, f"{self.ENDPOINT}/?country={country}"
        )

    def get_institution_by_id(self, id: str) -> Institutions:
        """
        Get details about specific institution by its id.

        Args:
            id (str): institution id (bank id)

        Returns:
            Institutions: Institutions json object
        """
        return self.__request(HTTPMethod.GET, f"{self.ENDPOINT}/{id}/")

    def get_institution_id_by_name(
        self, country: str, institution: str
    ) -> str:
        """
        Get institution id by institution name.

        Args:
            country (str): Two-character country code
            institution (str): Institution name (ex: Revolut)

        Raises:
            ValueError: If institution with given name is not found

        Returns:
            str: Institution id
        """
        institutions = self.get_institutions(country)

        for bank in institutions:
            if institution.lower() in bank["name"].lower():
                return bank["id"]

        raise ValueError(f"Institution: {institution} is not found")



from nordigen.types.http_enums import HTTPMethod
from nordigen.types.types import (
    AccountBalances,
    AccountData,
    AccountDetails,
    AgreementsList,
    EnduserAgreement,
    Institutions,
    Requisition,
    RequisitionDto,
    RequisitionList,
    TokenType,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nordigen', 'nordigen.api', 'nordigen.types']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.26.0,<3.0.0']

setup_kwargs = {
    'name': 'nordigen',
    'version': '1.0.1',
    'description': 'Python client for Nordigen API',
    'long_description': '# Nordigen Python\n\nThis is official Python client library for [Nordigen](https://nordigen.com/en) API\n\nFor a full list of endpoints and arguments, see the [docs](https://nordigen.com/en/account_information_documenation/api-documention/overview/).\n\nBefore starting to use API you will need to create a new secret and get your `SECRET_ID` and `SECRET_KEY` from the [Nordigen\'s Open Banking Portal](https://ob.nordigen.com/user-secrets/).\n\n## Requirements\n\n* Python >= 3.8\n\n\n## Installation\n\nInstall library via pip package manager:\n\n```\npip install nordigen\n```\n\n## Example application\n\nExample code can be found in `main.py` file and Flask application can be found in the `example` directory\n\n## Quickstart\n\n\n```python\nfrom uuid import uuid4\n\nfrom nordigen import NordigenClient\n\n# initialize Nordigen client and pass SECRET_ID and SECRET_KEY\nclient = NordigenClient(\n    secret_id="SECRET_ID",\n    secret_key="SECRET_KEY"\n)\n\n# Create new access and refresh token\n# Parameters can be loaded from .env or passed as a string\n# Note: access_token is automatically injected to other requests after you successfully obtain it\ntoken_data = client.generate_token()\n\n# Use existing token\nclient.token = "YOUR_TOKEN"\n\n# Exchange refresh token for new access token\nnew_token = client.exchange_token(token_data["refresh"])\n\n# Get institution id by bank name and country\ninstitution_id = client.institution.get_institution_id_by_name(\n    country="LV",\n    institution="Revolut"\n)\n\n# Get all institution by providing country code in ISO 3166 format\ninstitutions = client.institution.get_institutions("LV")\n\n# Initialize bank session\ninit = client.initialize_session(\n    # institution id\n    institution_id=institution_id,\n    # redirect url after successful authentication\n    redirect_uri="https://nordigen.com",\n    # additional layer of unique ID defined by you\n    reference_id=str(uuid4())\n)\n\n# Get requisition_id and link to initiate authorization process with a bank\nlink = init.link # bank authorization link\nrequisition_id = init.requisition_id\n```\n\nAfter successful authorization with a bank you can fetch your data (details, balances, transactions)\n\n---\n\n## Fetching account metadata, balances, details and transactions\n\n```python\n\n# Get account id after you have completed authorization with a bank\n# requisition_id can be gathered from initialize_session response\naccounts = client.requisition.get_requisition_by_id(\n    requisition_id=init.requisition_id\n)\n\n# Get account id from the list.\naccount_id = accounts["accounts"][0]\n\n# Create account instance and provide your account id from previous step\naccount = client.account_api(id=account_id)\n\n# Fetch account metadata\nmeta_data = account.get_metadata()\n# Fetch details\ndetails = account.get_details()\n# Fetch balances\nbalances = account.get_balances()\n# Fetch transactions\ntransactions = account.get_transactions()\n\n```\n\n## Support\n\nFor any inquiries please contact support at [support@nordigen.com](support@nordigen.com) or create an issue in repository.\n',
    'author': 'Nordigen Solutions',
    'author_email': 'support@nordigen.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nordigen/nordigen-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Final

from nordigen.types import AccountBalances
from nordigen.types.http_enums import HTTPMethod
from nordigen.types.types import AccountData, AccountDetails

if TYPE_CHECKING:
    from nordigen import NordigenClient


class AccountApi:
    """
    Account API endpoints to fetch accounts, details, balances and transactions.

    Attributes
    ---------
    client (NordigenClient):
        Injectable NordigenClient object to make an http requests
    id (str):
        account id retrieved from request get_requisition_by_id()

    Returns: None
    """

    __ENDPOINT: Final = "accounts"

    def __init__(self, client: NordigenClient, id: str) -> None:
        self.__request = client.request
        self.__id: str = id

    def __get(self, endpoint: str):
        """
        Construct get request.

        Args:
            endpoint (str): endpoint

        Returns:
            [type]: [description]
        """
        url = f"{self.__ENDPOINT}/{self.__id}/{endpoint}/"
        return self.__request(HTTPMethod.GET, f"{url}")

    def get_metadata(self) -> AccountData:
        """
        Access account metadata.
        Information about the account record, such as the processing status and IBAN.
        Account status is recalculated based on the error count in the latest req.

        Returns:
            AccountData: account metadata
        """
        return self.__request(
            HTTPMethod.GET, f"{self.__ENDPOINT}/{self.__id}/"
        )

    def get_balances(self) -> AccountBalances:
        """
        Access account balances.
        Balances will be returned in Berlin Group PSD2 format.

        Returns:
            AccountBalances: account balance data
        """
        return self.__get("balances")

    def get_details(self) -> AccountDetails:
        """
        Access account details.
        Account details will be returned in Berlin Group PSD2 format.

        Returns:
            AccountDetails: account details data
        """
        return self.__get("details")

    def get_transactions(self) -> Dict:
        """
        Access account transactions.
        Transactions will be returned in Berlin Group PSD2 format.


        Returns:
            Dict: account transactions details
        """
        return self.__get("transactions")



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Final, List

from nordigen.types.http_enums import HTTPMethod
from nordigen.types.types import Requisition

if TYPE_CHECKING:
    from nordigen import NordigenClient


class RequisitionsApi:
    """
    Requsition API class related to requisitions.

    Attributes
    ---------
    client(NordigenClient): Injectable NordigenClient object to make an http requests

    Returns: None
    """

    ENDPOINT: Final = "requisitions"

    def __init__(self, client: NordigenClient) -> None:
        self.__client = client

    def get_requisitions(
        self, limit: int = 100, offset: int = 0
    ) -> Requisition:
        """
        Get list of requisitions.

        Args:
            limit (int, optional): number of results to return per page. Defaults to 100.
            offset (int, optional): the initial index from which to return the results. Defaults to 0.

        Returns:
            Requisition: json response with requisition details
        """
        payload = {"limit": limit, "offset": offset}
        return self.__client.request(
            HTTPMethod.GET, f"{self.ENDPOINT}/", payload
        )

    def create_requisition(
        self,
        redirect_uri: str,
        reference_id: str,
        institution_id: str = None,
        agreement: List[str] = None,
        user_language: str = None,
    ) -> Requisition:
        """
        Create requisition for creating links and retrieving accounts.

        Args:
            redirect_uri (str): application redirect url
            reference_id (str): additional layer of unique ID defined by you
            enduser_id (str): a unique end-user ID of someone who's using your services, usually it's a UUID
            agreements (List[str] or str optional): agreement is provided as a string.
            user_language (str, optional): to enforce a language for all end user steps hosted
                by Nordigen passed as a two-letter country code. Defaults to None

        Returns:
            Requisition: [description]
        """
        payload = {
            "redirect": redirect_uri,
            "reference": reference_id,
            "institution_id": institution_id,
        }

        if user_language:
            payload["user_language"] = user_language

        if agreement:
            payload["agreement"] = agreement

        return self.__client.request(
            HTTPMethod.POST, f"{self.ENDPOINT}/", payload
        )

    def get_requisition_by_id(self, requisition_id: str) -> Requisition:
        """
        Get list of requisitions.

        Args:
            requisition_id (str): A UUID string identifying this requisition.
        Returns:
            Requisition: account details
        """
        return self.__client.request(
            HTTPMethod.GET, f"{self.ENDPOINT}/{requisition_id}/"
        )

    def delete_requisition(self, requisition_id: str) -> Dict:
        """
        Delete requisition by id.

        Args:
            requisition_id (str): A UUID string identifying this requisition.

        Returns: Dict that consist confirmation message that requisition has been deleted
        """
        return self.__client.request(
            HTTPMethod.DELETE, f"{self.ENDPOINT}/{requisition_id}"
        )
